import os
import requests
from http.server import BaseHTTPRequestHandler, HTTPServer
from urllib.parse import parse_qs
import traceback
import html
import re
from datetime import datetime

# === SECURITY WARNING ===
# Hardcoding API keys is insecure and can lead to unauthorized access.
# Ensure this script remains private and secure.
# ========================

OPENAI_API_KEY = "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"
PORT = 8003
MODEL = "gpt-3.5-turbo"
TEMPERATURE = 0.7

# Chat history stored in memory
CHAT_HISTORY = []

script_dir = os.path.dirname(os.path.abspath(__file__))
TEMPLATE_PATH = os.path.join(script_dir, "template.html")

try:
    with open(TEMPLATE_PATH, "r", encoding="utf-8") as f:
        HTML_TEMPLATE = f.read()
except Exception as e:
    print(f"Error reading template file at {TEMPLATE_PATH}: {e}")
    HTML_TEMPLATE = "<html><body><h1>Error loading template</h1></body></html>"

def format_code_blocks(content):
    pattern = re.compile(r'```(\w+)?\n(.*?)\n```', re.DOTALL)

    def replacer(match):
        language = match.group(1) if match.group(1) else ''
        code = html.escape(match.group(2))
        return f'<pre><code>{code}</code></pre>'

    return pattern.sub(replacer, content)

def generate_messages_html():
    html_parts = []
    message_pairs = []
    i = 0
    while i < len(CHAT_HISTORY):
        user_msg = CHAT_HISTORY[i]
        assistant_msg = CHAT_HISTORY[i + 1] if (i + 1) < len(CHAT_HISTORY) else None
        if user_msg.get("role") == "user" and assistant_msg and assistant_msg.get("role") == "assistant":
            message_pairs.append((user_msg, assistant_msg))
            i += 2
        else:
            message_pairs.append((user_msg, assistant_msg))
            i += 1

    message_pairs = list(reversed(message_pairs))

    for pair in message_pairs:
        user_msg, assistant_msg = pair
        if user_msg:
            role = user_msg.get("role", "")
            content = user_msg.get("content", "")
            timestamp = user_msg.get("timestamp", "")
            cls = "user-message"
            formatted_content = format_code_blocks(content)
            timestamp_str = f' <span class="timestamp">{timestamp}</span>' if timestamp else ''
            html_parts.append(f'<div class="message {cls}"><strong>{role.title()}:</strong><br>{formatted_content}{timestamp_str}</div>')

        if assistant_msg:
            role = assistant_msg.get("role", "")
            content = assistant_msg.get("content", "")
            timestamp = assistant_msg.get("timestamp", "")
            cls = "assistant-message"
            formatted_content = format_code_blocks(content)
            timestamp_str = f' <span class="timestamp">{timestamp}</span>' if timestamp else ''
            html_parts.append(f'<div class="message {cls}"><strong>{role.title()}:</strong><br>{formatted_content}{timestamp_str}</div>')

    return "\n".join(html_parts)

def openai_request(prompt):
    req_data = {
        "model": MODEL,
        "messages": [{"role": "system", "content": "You are a helpful assistant."}] + CHAT_HISTORY + [{"role": "user", "content": prompt}],
        "temperature": TEMPERATURE,
        "stream": False
    }
    try:
        response = requests.post(
            "https://api.openai.com/v1/chat/completions",
            headers={
                "Content-Type": "application/json",
                "Authorization": f"Bearer {OPENAI_API_KEY}"
            },
            json=req_data,
            timeout=30
        )

        if response.status_code != 200:
            print(f"HTTP Error during OpenAI request: {response.status_code} - {response.text}")
            return f"Error: Unable to fetch response from OpenAI (HTTP {response.status_code})."

        resp_data = response.json()
        print("OpenAI API Response:", resp_data)
        choices = resp_data.get("choices", [])
        if not choices:
            return "Error: No response from OpenAI."
        message = choices[0].get("message", {})
        content = message.get("content", "")
        return content
    except Exception as e:
        print(f"Error during OpenAI request: {e}")
        traceback.print_exc()
        return "Error: Unable to connect to OpenAI."

class Handler(BaseHTTPRequestHandler):
    def do_GET(self):
        try:
            messages_html = generate_messages_html()
            page = HTML_TEMPLATE.replace('{{ messages }}', messages_html)
            self.send_response(200)
            self.send_header("Content-type", "text/html; charset=utf-8")
            self.end_headers()
            self.wfile.write(page.encode("utf-8"))
        except Exception as e:
            self.send_error(500, f"Internal Server Error: {e}")
            traceback.print_exc()

    def do_POST(self):
        try:
            length = int(self.headers.get('Content-Length', 0))
            body = self.rfile.read(length).decode('utf-8')
            params = parse_qs(body)

            action = params.get("action", [None])[0]
            question = params.get("question", [""])[0].strip()

            if action == "clear_history":
                CHAT_HISTORY.clear()
                print("Chat history cleared.")
            elif question:
                if len(question) > 500:
                    raise ValueError("Input too long. Maximum 500 characters allowed.")
                timestamp = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
                CHAT_HISTORY.append({"role": "user", "content": question, "timestamp": timestamp})
                answer = openai_request(question)
                timestamp = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
                CHAT_HISTORY.append({"role": "assistant", "content": answer, "timestamp": timestamp})
            else:
                print("No valid action or question provided.")

            messages_html = generate_messages_html()
            page = HTML_TEMPLATE.replace('{{ messages }}', messages_html)
            self.send_response(200)
            self.send_header("Content-type", "text/html; charset=utf-8")
            self.end_headers()
            self.wfile.write(page.encode("utf-8"))
        except ValueError as ve:
            self.send_error(400, f"Bad Request: {ve}")
            print(f"ValueError: {ve}")
        except Exception as e:
            self.send_error(500, f"Internal Server Error: {e}")
            traceback.print_exc()

    def log_message(self, format, *args):
        return

if __name__ == "__main__":
    try:
        server_address = ('0.0.0.0', PORT)
        server = HTTPServer(server_address, Handler)
        print(f"Serving on port {PORT}")
        server.serve_forever()
    except KeyboardInterrupt:
        print("\nServer shutting down.")
    except Exception as e:
        print(f"Error starting server: {e}")
        traceback.print_exc()
